package dev.pfaff.altargui;

import dev.pfaff.altargui.element.HasDebugLayer;
import dev.pfaff.altargui.element.Painter;
import dev.pfaff.altargui.element.PostPainter;
import dev.pfaff.altargui.handlers.MouseDragHandler;
import dev.pfaff.altargui.handlers.MouseMoveHandler;
import dev.pfaff.altargui.handlers.MouseScrollHandler;
import dev.pfaff.altargui.inspector.DebugLayer;
import dev.pfaff.altargui.inspector.InspectorMeta;
import dev.pfaff.altargui.internal.util.Specialize;
import dev.pfaff.altargui.render.objects.DebugRectRenderObject;
import dev.pfaff.altargui.selector.BuiltinSelector;
import dev.pfaff.altargui.util.CoordUtil;
import dev.pfaff.altargui.util.Preconditions;
import dev.pfaff.altargui.util.listenable.ListenableFlag;
import dev.pfaff.altargui.util.listenable.ListenerDisposer;
import dev.pfaff.altargui.widgets.IdWidget;
import org.jetbrains.annotations.MustBeInvokedByOverriders;
import org.jetbrains.annotations.Nullable;

import java.util.List;
import java.util.function.BiConsumer;
import java.util.function.Consumer;

import static dev.pfaff.altargui.GuiGlobals.CHECKS;
import static dev.pfaff.altargui.GuiGlobals.LOGGER;
import static dev.pfaff.log4truth.StandardTags.WARN;
import static java.util.FormatProcessor.FMT;

public non-sealed abstract class WidgetInstance<W extends Widget> implements IElement {
	public static final WidgetInstance[] EMPTY_ARRAY = new WidgetInstance[0];
	static final boolean CHECK_STATE = false;

	/**
	 * The cached key of the element's widget. This <strong>must</strong> be the first field in the class for offset
	 * compatibility with {@link GuiContext.RootElement}.
	 */
	final long key;
	final GuiContext guiContext;
	@InspectorMeta(tooltip = "The class of the element's Widget")
	final Class<? extends Widget> widgetClass;
	@InspectorMeta(tooltip = "The unique identifier of this element")
	boolean generation;
	// @Stable
	boolean initialized;

	@Nullable
	IElement parent;

	final @Nullable DebugLayer debugLayer = this instanceof HasDebugLayer hasDebugLayer
			? hasDebugLayer.debugLayer()
			: null;
	final @Nullable DebugRectRenderObject debugRect = this instanceof HasDebugLayer hasDebugLayer
			? new DebugRectRenderObject()
			: null;

	final @Nullable MouseDragHandler mouseDragHandler = this instanceof MouseDragHandler handler ? handler : null;
	final @Nullable MouseMoveHandler mouseMoveHandler = this instanceof MouseMoveHandler handler ? handler : null;
	final @Nullable MouseScrollHandler mouseScrollHandler = this instanceof MouseScrollHandler handler ? handler : null;
	final @Nullable Painter painter = this instanceof Painter painter ? painter : null;
	final @Nullable PostPainter postPainter = this instanceof PostPainter painter ? painter : null;

	protected WidgetInstance(WidgetInstantiationArgs args) {
		this.guiContext = args.context;
		this.widgetClass = args.widgetClass;
		this.key = args.key;
	}

	@InspectorMeta(tooltip = "The X-axis position of the widget")
	private float allocatedX;
	@InspectorMeta(tooltip = "The Y-axis position of the widget")
	private float allocatedY;
	@InspectorMeta(tooltip = "The allocated width of the widget")
	private float allocatedWidth;
	@InspectorMeta(tooltip = "The allocated height of the widget")
	private float allocatedHeight;
	@InspectorMeta(tooltip = "Whether the widget is permitted by the parent to overflow the X-axis")
	private boolean canOverflowX;
	@InspectorMeta(tooltip = "Whether the widget is permitted by the parent to overflow the Y-axis")
	private boolean canOverflowY;
	@InspectorMeta(tooltip = "The computed width of the widget")
	private float computedWidth;
	@InspectorMeta(tooltip = "The computed height of the widget")
	private float computedHeight;

	// @Stable
	private boolean hasStencilLaidOut = false;
	private boolean hasTraceLaidOut = false;
	private boolean isStencilLayingOut = false;
	private boolean isTraceLayingOut = false;
	// @Stable
	private boolean isDisposed = false;
	public final ListenableFlag needsStencilLayout = new ListenableFlag(true);
	public final ListenableFlag needsTraceLayout = new ListenableFlag(true);
	private @Nullable ListenerDisposer needsStencilLayoutParentNotifier;
	private @Nullable ListenerDisposer needsTraceLayoutParentNotifier;

	final boolean isDisposed() {
		return this.isDisposed;
	}

	/**
	 * Configures the instance's state according to the inputs in the specified {@code widget}.
	 * <p>
	 * Other initialization might be done if {@code init} is true.
	 */
	// TODO: explore whether moving this to the Widget interface would improve performance.
	public abstract void configure(StateContext stateContext, W widget, boolean init);

	/**
	 * Called when the element is mounted to a new parent.
	 */
	public void onMount() {
		if (CHECK_STATE && this.isDisposed) throw new IllegalStateException("Already disposed");
		// add listener to propagate needs layout upwards
		if (parent instanceof WidgetInstance<?> parentWidgetInstance) {
			needsStencilLayoutParentNotifier = needsStencilLayout.addListener(parentWidgetInstance::markNeedsStencilLayout);
			needsTraceLayoutParentNotifier = needsTraceLayout.addListener(parentWidgetInstance::markNeedsTraceLayout);
		}
	}

	/**
	 * Called when the element is unmounted from the current parent.
	 */
	@MustBeInvokedByOverriders
	public void onUnmount() {
		if (CHECK_STATE && this.isDisposed) throw new IllegalStateException("Already disposed");
		var needsStencilLayoutParentNotifier = this.needsStencilLayoutParentNotifier;
		if (needsStencilLayoutParentNotifier != null) {
			needsStencilLayoutParentNotifier.close();
			this.needsStencilLayoutParentNotifier = null;
		}
		var needsTraceLayoutParentNotifier = this.needsTraceLayoutParentNotifier;
		if (needsTraceLayoutParentNotifier != null) {
			needsTraceLayoutParentNotifier.close();
			this.needsTraceLayoutParentNotifier = null;
		}
	}

	/**
	 * Disposes of the element... currently the element may be reused, but this will likely change in the future.
	 */
	@MustBeInvokedByOverriders
	public void dispose() {
		if (CHECK_STATE && this.isDisposed) throw new IllegalStateException("Already disposed");
		onUnmount();
		parent = null;
		if (CHECK_STATE) this.isDisposed = true;
	}

	/**
	 * Sets positions of element content and children.
	 * <p>
	 * This will usually call {@link StencilLayoutContext#layout(WidgetInstance, WidgetInstance, BiConsumer)}
	 * with each child and go through a layout procedure wherein
	 * {@link WidgetInstance#setAvailableSize(float, float)} and then
	 * {@link WidgetInstance#stencilLayout(StencilLayoutContext)} are called (in that order) on each child.
	 */
	public final void stencilLayout(StencilLayoutContext context) {
		if (CHECKS && (isStencilLayingOut || isTraceLayingOut)) {
			throw new IllegalStateException("Already laying out: " + this);
		}
		if (!takeNeedsStencilLayout()) {
			if (CHECKS && !hasStencilLaidOut) {
				throw new IllegalStateException("Not marked for layout, but never been laid out: " + this);
			}
			return;
		}
		if (CHECKS) {
			isStencilLayingOut = true;
		}
//		LOGGER.log(Level.INFO, FMT."Took needs layout on widget@%016x\{key}, a \{widgetClass.getName()}");
		computeStencilLayout(context);
		if (CHECKS) {
			if (!hasStencilLaidOut) hasStencilLaidOut = true;
			isStencilLayingOut = false;
		}
	}

	/**
	 * mounts children and computes actual size and position from the allocated rect
	 * <p>
	 * This will usually call {@link StencilLayoutContext#layout(WidgetInstance, WidgetInstance, BiConsumer)}
	 * with each child and go through a layout procedure wherein
	 * {@link WidgetInstance#setAvailableSize(float, float)} and then
	 * {@link WidgetInstance#traceLayout()} are called (in that order) on each child.
	 */
	public final void traceLayout() {
		try {
			if (CHECKS && (isTraceLayingOut || isStencilLayingOut)) {
				throw new IllegalStateException("Already laying out: " + this);
			}
			if (!takeNeedsTraceLayout()) {
				if (CHECKS && !hasTraceLaidOut) {
					throw new IllegalStateException("Not marked for layout, but never been laid out: " + this);
				}
				return;
			}
			if (CHECKS) {
				isTraceLayingOut = true;
			}
	//		LOGGER.log(Level.INFO, FMT."Took needs layout on widget@%016x\{key}, a \{widgetClass.getName()}");
			computeTraceLayout();
			if (CHECKS) {
				if (!hasTraceLaidOut) hasTraceLaidOut = true;
				isTraceLayingOut = false;
			}
		} catch (Throwable e) {
			throw GuiException.wrap(FMT."Laying out \{this.toStringShort()}", e);
		}
	}

	public final void mount(MountContext context) {
		if (CHECKS && (!hasStencilLaidOut || !hasTraceLaidOut)) {
			throw new IllegalStateException("Attempted to mount an element that has never been laid out: " + this);
		}
		computeMount(context);
	}

	@Specialize
	public void uSpecMount(MountContext context) {
		context.iSpecMount(this);
	}

	protected static void computeFillLayout(WidgetInstance element, StencilLayoutContext context) {
		element.requireFiniteAllocatedWidth();
		element.requireFiniteAllocatedHeight();
		element.fillAllocated();
	}

	protected abstract void computeStencilLayout(StencilLayoutContext context);

	protected abstract void computeTraceLayout();

	protected abstract void computeMount(MountContext context);

	protected final void markNeedsStencilLayout() {
//		if (needsLayout.value()) {
//			LOGGER.log(Level.INFO, FMT."Marked needs layout on widget@%016x\{key}, a \{widgetClass.getName()}");
//		}
		needsStencilLayout.set();
	}

	protected final void markNeedsTraceLayout() {
//		if (needsLayout.value()) {
//			LOGGER.log(Level.INFO, FMT."Marked needs layout on widget@%016x\{key}, a \{widgetClass.getName()}");
//		}
		needsTraceLayout.set();
	}

	/**
	 * @deprecated use {@link #markNeedsStencilLayout()} and/or {@link #markNeedsTraceLayout()} as appropriate instead.
	 */
	@Deprecated
	protected final void markNeedsLayout() {
		markNeedsStencilLayout();
		markNeedsTraceLayout();
	}

	/**
	 * This will need to be overridden for any aligned positioning to include {@link #markNeedsTraceLayout()}.
	 */
	protected void propagateNeedsStencilLayoutFromChild() {
		markNeedsStencilLayout();
	}

	protected final boolean takeNeedsStencilLayout() {
		return needsStencilLayout.takeNoNotify();
	}

	protected final boolean takeNeedsTraceLayout() {
		return needsTraceLayout.takeNoNotify();
	}

	public final float allocatedX() {
		if (CHECKS && isStencilLayingOut) {
			throw new IllegalStateException("Attempted to access allocatedX while stencil laying out: " + toStringShort());
		}
		return allocatedX;
	}

	public final float allocatedY() {
		if (CHECKS && isStencilLayingOut) {
			throw new IllegalStateException("Attempted to access allocatedY while stencil laying out: " + toStringShort());
		}
		return allocatedY;
	}

	public final float allocatedWidth() {
		return allocatedWidth;
	}

	public final float allocatedHeight() {
		return allocatedHeight;
	}

	protected final void markNeedsStencilLayoutOnAllocationChange() {
//		markNeedsStencilLayout();
		needsStencilLayout.setNoNotify();
	}

	protected final void markNeedsTraceLayoutOnAllocationChange() {
//		markNeedsTraceLayout();
		needsTraceLayout.setNoNotify();
	}

	/**
	 * Called by the parent element to set the position of its children.
	 */
	public final void setPosition(float x, float y) {
		if (this.allocatedX != x || this.allocatedY != y) {
			// FIXME: this will notify the parent again, causing an unnecessary layout. We cannot simply skip notifying,
			//  because there may be other listeners that actually require the notification.
			markNeedsTraceLayoutOnAllocationChange();
			checkFinite(x, "allocatedX");
			checkFinite(y, "allocatedY");
			this.allocatedX = x;
			this.allocatedY = y;
		}
	}

	/**
	 * Called by the parent element to communicate the amount of space available to its children.
	 */
	public final void setAvailableSize(float width, float height) {
		if (this.allocatedWidth != width || this.allocatedHeight != height) {
			// FIXME: this will notify the parent again, causing an unnecessary layout. We cannot simply skip notifying,
			//  because there may be other listeners that actually require the notification.
			markNeedsStencilLayoutOnAllocationChange();
			checkFinite(width, "allocatedWidth");
			checkFinite(height, "allocatedHeight");
			this.allocatedWidth = Math.max(width, 0f);
			this.allocatedHeight = Math.max(height, 0f);
		}
	}

	/**
	 * Copies the {@linkplain #setPosition(float, float) position} from the given {@code other} widget (usually the
	 * parent).
	 */
	public final void copyPositionFrom(WidgetInstance other) {
		setPosition(other.allocatedX, other.allocatedY);
	}

	/**
	 * Copies the {@linkplain #setAvailableSize(float, float) available size} from the given {@code other} widget
	 * (usually the parent).
	 */
	public final void copyAvailableSizeFrom(WidgetInstance other) {
		setAvailableSize(other.allocatedWidth, other.allocatedHeight);
	}

	protected final void requireFiniteAllocatedWidth() {
		if (allocatedWidth() == Float.POSITIVE_INFINITY) {
			throw new GuiException(getClass().getSimpleName() + " cannot be used with infinite allocated width");
		}
	}

	protected final void requireFiniteAllocatedHeight() {
		if (allocatedHeight() == Float.POSITIVE_INFINITY) {
			throw new GuiException(getClass().getSimpleName() + " cannot be used with infinite allocated height");
		}
	}

	public final float computedWidth() {
		return computedWidth;
	}

	public final float computedHeight() {
		return computedHeight;
	}

	protected final void setComputedSize(float width, float height) {
		if (CHECKS && !isStencilLayingOut) {
			throw new IllegalStateException("Not stencil laying out: " + this.toStringShort());
		}
		if (computedWidth != width || computedHeight != height) {
			checkFinite(width, "computedWidth");
			checkPositive(width, "computedWidth");
			checkFinite(height, "computedHeight");
			checkPositive(height, "computedHeight");
			computedWidth = width;
			computedHeight = height;
		}
	}

	private void checkFieldsEqual(float ours, float theirs, String name) {
		if (CHECKS) {
			Preconditions.checkArgument(theirs == ours,
										() -> STR."theirs.\{name} (\{theirs}) != ours.\{name} (\{ours})");
		}
	}

	protected final void copyComputedSizeFrom(WidgetInstance other) {
		setComputedSize(other.computedWidth(), other.computedHeight());
	}

	/**
	 * Sets the computed size to fill the entire allocated rect.
	 */
	protected final void fillAllocated() {
		setComputedSize(allocatedWidth(), allocatedHeight());
	}

	public final boolean isInside(double checkX, double checkY) {
		return CoordUtil.isInside(allocatedX,
								  allocatedY,
								  computedWidth,
								  computedHeight,
								  checkX,
								  checkY);
	}

	public void applyBuiltinSelector(BuiltinSelector selector, Consumer<WidgetInstance> consumer) {
		LOGGER.log(() -> STR."attempted to apply an unsupported builtin selector (\{selector}) on \{this}", List.of(WARN));
	}

	public final String toStringShort() {
		if (this instanceof IdWidget.State idWidget) {
			return FMT."\{this.widgetClass.getName()}@%016x\{this.key} #\{idWidget.id()}";
		}
		return FMT."\{this.widgetClass.getName()}@%016x\{this.key}";
	}

	@Override
	public final String toString() {
		return FMT."""
		WidgetInstance[
		  class=\{this.widgetClass.getName()}
		  key=%016x\{this.key}
		  generation=\{this.generation}
		  initialized=\{this.initialized}
		]""";
	}

	private static void checkFinite(float f, String name) {
		if (!Float.isFinite(f)) {
			throw new IllegalArgumentException(STR."Expected a finite value for argument '\{name}', found \{f}");
		}
	}

	private static void checkPositive(float f, String name) {
		if (f < 0) {
			throw new IllegalArgumentException(STR."Expected a positive value for argument '\{name}', found \{f}");
		}
	}

	public final GuiContext guiContext() {
		return guiContext;
	}

	@Override
	public final long key() {
		return key;
	}

	@Override
	public final boolean equals(Object o) {
		return super.equals(o);
	}

	@Override
	public final int hashCode() {
		return super.hashCode();
	}
}
