package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.inspector.DebugLayer;
import dev.pfaff.altargui.selector.BuiltinSelector;
import dev.pfaff.altargui.selector.Children;
import dev.pfaff.altargui.selector.NthElement;
import dev.pfaff.altargui.util.ListMultiChildBuilder;
import dev.pfaff.altargui.util.NonNullLong2WidgetInstanceHashMap;
import org.jetbrains.annotations.NotNull;
import org.jspecify.annotations.NonNull;

import java.util.List;
import java.util.function.BiConsumer;
import java.util.function.Consumer;

/**
 * Builder is only <em>guaranteed</em> to be called when it changes.
 */
public record MultiChildWidget(@NotNull Mode mode, Builder builder) implements Widget {
	public static final DebugLayer DEBUG_LAYER = new DebugLayer(MultiChildWidget.class);

	private static final MultiChildWidget EMPTY_STACK = new MultiChildWidget(Mode.Stack, ignored -> {});
	private static final MultiChildWidget EMPTY_COL = new MultiChildWidget(Mode.Column, ignored -> {});
	private static final MultiChildWidget EMPTY_ROW = new MultiChildWidget(Mode.Row, ignored -> {});
	private static final MultiChildWidget EMPTY_DIAGONAL = new MultiChildWidget(Mode.Diagonal, ignored -> {});

	public enum Mode {
		Stack(true, true),
		Row(false, true),
		Column(true, false),
		Diagonal(false, false);

		final boolean stackX;
		final boolean stackY;

		Mode(boolean stackX, boolean stackY) {
			this.stackX = stackX;
			this.stackY = stackY;
		}
	}

	public static MultiChildWidget stack(Builder builder) {
		return new MultiChildWidget(Mode.Stack, builder);
	}

	public static MultiChildWidget row(Builder builder) {
		return new MultiChildWidget(Mode.Row, builder);
	}

	public static MultiChildWidget col(Builder builder) {
		return new MultiChildWidget(Mode.Column, builder);
	}

	public static MultiChildWidget diagonal(Builder builder) {
		return new MultiChildWidget(Mode.Diagonal, builder);
	}

	public static MultiChildWidget stack(List<? extends Widget> children) {
		return stack(ListMultiChildBuilder.newUntrusted(children));
	}

	public static MultiChildWidget row(List<? extends Widget> children) {
		return row(ListMultiChildBuilder.newUntrusted(children));
	}

	public static MultiChildWidget col(List<? extends Widget> children) {
		return col(ListMultiChildBuilder.newUntrusted(children));
	}

	public static MultiChildWidget diagonal(List<? extends Widget> children) {
		return diagonal(ListMultiChildBuilder.newUntrusted(children));
	}

	public static MultiChildWidget emptyStack() {
		return EMPTY_STACK;
	}

	public static MultiChildWidget emptyCol() {
		return EMPTY_COL;
	}

	public static MultiChildWidget emptyRow() {
		return EMPTY_ROW;
	}

	public static MultiChildWidget emptyDiagonal() {
		return EMPTY_DIAGONAL;
	}

	@Override
	public @NotNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	@FunctionalInterface
	public interface Builder extends Consumer<ManagedLayoutContext> {
		@Override
		void accept(ManagedLayoutContext context);

		@NotNull
		@Override
		default Builder andThen(@NotNull Consumer<? super ManagedLayoutContext> after) {
			return context -> {
				this.accept(context);
				after.accept(context);
			};
		}
	}

	private static final class State extends WidgetInstance<MultiChildWidget> {
		private @NonNull Mode mode;
		private Builder builder;
		private final WidgetInstanceList children = new WidgetInstanceList(2);

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, MultiChildWidget widget, boolean init) {
			mode = widget.mode;
			if (!widget.builder.equals(builder)) {
				builder = widget.builder;
				children.clear();
				builder.accept((key, widget1) -> {
					children.add(stateContext.getState(this, key, widget1));
				});
				markNeedsStencilLayout();
				markNeedsTraceLayout();
			}
		}

		@Override
		protected void computeStencilLayout(StencilLayoutContext context) {
			var children = this.children;
			var mode = this.mode;
			final float allocWidth = this.allocatedWidth();
			final var allocHeight = this.allocatedHeight();

			var layoutFunction = new NonTightLayoutFunction(allocWidth, allocHeight, mode);
			children.forEach(child -> context.layout(this, child, layoutFunction));
			if (layoutFunction.anyTight) {
				children.forEach(child -> {
					if (child instanceof SizedWidget.TightElement tightChild) {
						if (tightChild.isX() && !mode.stackX) {
							throw makeCannotLayoutTightChildException(mode, "width");
						}
						if (tightChild.isY() && !mode.stackY) {
							throw makeCannotLayoutTightChildException(mode, "height");
						}
						float width_ = tightChild.isX() ? layoutFunction.maxWidth : layoutFunction.remainingWidth;
						float height_ = tightChild.isY() ? layoutFunction.maxHeight : layoutFunction.remainingHeight;
						var tightLayoutFunction = new TightLayoutFunction(width_, height_, mode);
						tightLayoutFunction.accept((WidgetInstance<?>) tightChild, context);
						if (!tightChild.isX()) {
							layoutFunction.remainingWidth = tightLayoutFunction.remainingWidth;
						}
						if (!tightChild.isY()) {
							layoutFunction.remainingHeight = tightLayoutFunction.remainingHeight;
						}
						layoutFunction.maxWidth = Math.max(layoutFunction.maxWidth,
								tightLayoutFunction.maxWidth);
						layoutFunction.maxHeight = Math.max(layoutFunction.maxHeight,
								tightLayoutFunction.maxHeight);
					}
				});
			}
			setComputedSize(mode.stackX ? layoutFunction.maxWidth : allocWidth - layoutFunction.remainingWidth,
					mode.stackY ? layoutFunction.maxHeight : allocHeight - layoutFunction.remainingHeight);
		}

		private static GuiException makeCannotLayoutTightChildException(Mode mode, String property) {
			return new GuiException(STR."Cannot layout a child with a tight \{property} in a \{mode} MultiChildWidget");
		}

		@Override
		protected void computeTraceLayout() {
			var children = this.children;
			var mode = this.mode;
			final float minX = this.allocatedX();
			final float minY = this.allocatedY();
			switch (mode) {
				case Stack -> {
					// TODO: is is more efficient to do these in the same for loop?
					children.forEach(child -> child.setPosition(minX, minY));
					children.forEach(WidgetInstance::traceLayout);
				}
				case Row -> children.forEach(new NonNullLong2WidgetInstanceHashMap.WidgetInstanceConsumer() {
					private float offX = 0f;

					@Override
					public void accept(WidgetInstance<?> child) {
						child.setPosition(minX + this.offX, minY);
						this.offX += child.computedWidth();
						child.traceLayout();
					}
				});
				case Column -> children.forEach(new NonNullLong2WidgetInstanceHashMap.WidgetInstanceConsumer() {
					private float offY = 0f;

					@Override
					public void accept(WidgetInstance<?> child) {
						child.setPosition(minX, minY + this.offY);
						this.offY += child.computedHeight();
						child.traceLayout();
					}
				});
				case Diagonal -> children.forEach(new NonNullLong2WidgetInstanceHashMap.WidgetInstanceConsumer() {
					private float offX = 0f;
					private float offY = 0f;

					@Override
					public void accept(WidgetInstance<?> child) {
						child.setPosition(minX + this.offX, minY + this.offY);
						this.offX += child.computedWidth();
						this.offY += child.computedHeight();
						child.traceLayout();
					}
				});
			}
		}

		@Override
		protected void computeMount(MountContext context) {
			context.mountList(children);
		}

		@Override
		public void uSpecMount(MountContext context) {
			super.uSpecMount(context);
		}

		@Override
		public void applyBuiltinSelector(BuiltinSelector selector, Consumer<WidgetInstance> consumer) {
			switch (selector) {
				case NthElement nthElement -> {
					int n = nthElement.n();
					if (n < 0) n = this.children.size() - n;
					if (n >= 0 && n < this.children.size()) {
						BuiltinSelector.applyNext(this.children.get(n),
												  nthElement,
												  consumer);
					}
				}
				case Children children -> {
					this.children.forEach(child -> BuiltinSelector.applyNext(child, children, consumer));
				}
			}
		}
	}

	private static abstract class LayoutFunction implements BiConsumer<WidgetInstance<?>, StencilLayoutContext> {
		private final boolean stackX;
		private final boolean stackY;

		float remainingWidth;
		float remainingHeight;
		float maxWidth = 0f;
		float maxHeight = 0f;

		public LayoutFunction(float width, float height, Mode mode) {
			this.stackX = mode.stackX;
			this.stackY = mode.stackY;

			this.remainingWidth = width;
			this.remainingHeight = height;
		}

		protected abstract boolean preAccept(WidgetInstance<?> e);

		@Override
		public final void accept(WidgetInstance e, StencilLayoutContext ctx) {
			if (!preAccept(e)) return;

			e.setAvailableSize(this.remainingWidth, this.remainingHeight);
			e.stencilLayout(ctx);

			var computedWidth = e.computedWidth();
			var computedHeight = e.computedHeight();

			if (stackX) {
				this.maxWidth = Math.max(this.maxWidth, computedWidth);
			} else {
				this.remainingWidth -= computedWidth;
			}
			if (stackY) {
				this.maxHeight = Math.max(this.maxHeight, computedHeight);
			} else {
				this.remainingHeight -= computedHeight;
			}
		}
	}

	private static final class NonTightLayoutFunction extends LayoutFunction {
		private boolean anyTight = false;

		public NonTightLayoutFunction(float width, float height, Mode mode) {
			super(width, height, mode);
		}

		@Override
		protected boolean preAccept(WidgetInstance<?> e) {
			if (e instanceof SizedWidget.TightElement) {
				this.anyTight = true;
				return false;
			}
			return true;
		}
	}

	private static final class TightLayoutFunction extends LayoutFunction {
		public TightLayoutFunction(float width, float height, Mode mode) {
			super(width, height, mode);
		}

		@Override
		protected boolean preAccept(WidgetInstance<?> e) {
			var tightE = (SizedWidget.TightElement) e;
			return true;
		}
	}
}
