package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.types.Alignment;
import org.jetbrains.annotations.NotNull;

public sealed interface PositionedWidget extends SingleChildWidget {
	public record Absolute(float x, float y, Widget child) implements PositionedWidget {
		@Override
		public @NotNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
			return new State(args);
		}

		private static final class State extends CustomLayoutState<Absolute> {
			private float x;
			private float y;

			private State(WidgetInstantiationArgs args) {
				super(args);
			}

			@Override
			public void configure(StateContext stateContext, Absolute widget, boolean init) {
				super.configure(stateContext, widget, init);
				if (x != widget.x || y != widget.y) {
					x = widget.x;
					y = widget.y;
					markNeedsLayout();
				}
			}

			@Override
			protected void computeStencilLayout(StencilLayoutContext context) {
				context.layout(this, getChildElement(), (element, ctx) -> {
//					var allocWidth = mcWindow.getScaledWidth();
//					var allocHeight = mcWindow.getScaledHeight();
					var allocWidth = this.allocatedWidth();
					var allocHeight = this.allocatedHeight();
					element.setAvailableSize(allocWidth - this.x, allocHeight - this.y);
					element.stencilLayout(ctx);
					setComputedSize(0, 0);
				});
			}

			@Override
			protected void computeTraceLayout() {
				var e = getChildElement();
				e.setPosition(this.x, this.y);
				e.traceLayout();
			}
		}
	}

	public record Relative(float x, float y, Widget child) implements PositionedWidget {
		@Override
		public @NotNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
			return new State(args);
		}

		private static final class State extends CustomLayoutState<Relative> {
			private float x;
			private float y;

			private State(WidgetInstantiationArgs args) {
				super(args);
			}

			@Override
			public void configure(StateContext stateContext, Relative widget, boolean init) {
				super.configure(stateContext, widget, init);
				if (x != widget.x || y != widget.y) {
					x = widget.x;
					y = widget.y;
					markNeedsLayout();
				}
			}

			@Override
			protected void computeStencilLayout(StencilLayoutContext context) {
				context.layout(this, getChildElement(), (element, ctx) -> {
					element.setAvailableSize(allocatedWidth() - x, allocatedHeight() - y);
					element.stencilLayout(ctx);
					setComputedSize(Math.max(0f, x + element.computedWidth()), Math.max(0f, y + element.computedHeight()));
				});
			}

			@Override
			protected void computeTraceLayout() {
				var e = getChildElement();
				e.setPosition(allocatedX() + x, allocatedY() + y);
				e.traceLayout();
			}
		}
	}

	public record Aligned(Widget child, Alignment horizontal, Alignment vertical) implements PositionedWidget {
		@Override
		public @NotNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
			return new State(args);
		}

		private static final class State extends CustomLayoutState<Aligned> {
			private @NotNull Alignment horizontal;
			private @NotNull Alignment vertical;

			private float childOffsetX;
			private float childOffsetY;

			private State(WidgetInstantiationArgs args) {
				super(args);
			}

			@Override
			public void configure(StateContext stateContext, Aligned widget, boolean init) {
				super.configure(stateContext, widget, init);
				if (horizontal != widget.horizontal || vertical != widget.vertical) {
					horizontal = widget.horizontal;
					vertical = widget.vertical;
					markNeedsLayout();
				}
			}

			@Override
			protected void computeStencilLayout(StencilLayoutContext context) {
				var horizontal = this.horizontal;
				var vertical = this.vertical;
				context.layout(this, getChildElement(), (e, ctx) -> {
					e.copyAvailableSizeFrom(this);
					e.stencilLayout(ctx);
					var allocWidth = allocatedWidth();
					var allocHeight = allocatedHeight();
					this.childOffsetX = horizontal.apply(e.computedWidth(), allocWidth);
					this.childOffsetY = vertical.apply(e.computedHeight(), allocHeight);
					setComputedSize(switch (horizontal) {
						case Start -> this.childOffsetX + e.computedWidth();
						case Center, End -> allocWidth;
					}, switch (vertical) {
						case Start -> this.childOffsetY + e.computedHeight();
						case Center, End -> allocHeight;
					});
				});
			}

			@Override
			protected void computeTraceLayout() {
				var e = getChildElement();
				e.setPosition(allocatedX() + this.childOffsetX, allocatedY() + this.childOffsetY);
				e.traceLayout();
			}
		}
	}
}
